!-----------------------------------------------------------------------------
! Magic And Fight System --- Magic Device
! For INFORM compiler and Z-Machine text based Adventure games
!-----------------------------------------------------------------------------
!  Author: Giancarlo Niccolai
!  Version: 1.3
!  Last Revised: 2002-11-24
!  Contact author at: giancarlo@niccolai.ws
!  Feel free to use and distribute this file, but please report me if
!  you intend to change it (in better), so we can upgrade together
!
!  Released under GPL 2.0, or any later version at your will
!
!-----------------------------------------------------------------------------
! Magic devices are objects that can provide means to cast spells other than
! the caster memory. Objects that do not require cooperation of the caster
! memory can be created with this class.
!-----------------------------------------------------------------------------

Ifndef MagicDevice;

Ifndef Spell;
	include "MFS_Magic.h";
Endif;

class MagicDevice
	with
	charges -1,					! infinite charges
	verbosity MFS_SPELLVMED,	! Medium verbosity in description

	! Describes the object: to be used as superclass routine
	! from subclasses;
	description [;
		if (self hasnt known_about) return;
		self.describe_contents();
		switch(self.verbosity) {
			MFS_SPELLVLO: rfalse;
			MFS_SPELLVMED:
				switch (self.charges) {
					-1: print (The) self, " sembra avere poteri intatti.^";
					0: print (The) self, " sembra aver esaurito tutti i suoi poteri.^";
					1: print (The) self, " sembra avere ancora poco potere rimasto.^";
					2: print (The) self, " sembra avere ancora un certo potere.^";
					3 to 999: print (The) self, " sembra potentissimo.^";
				}
				rtrue;

			MFS_SPELLVHI:
				if (self.charges == -1)
					"Puoi usare ", (the) self, " quandunque tu voglia.^";
				else {
					print "Puoi usare ", (the) self, " ";
					print (number) self.charges, " volt", (numf) self.charges, ".^";
				}
			rtrue;
		}
	],

	affect [ spell;
		if (spell ~= SPELL_RECHARGE) rfalse;

		self.charges = self.max_charges;
		"Mentre il potere della magia riempie ", (the) self,
		", ", (thatorthose) self, " inizia a brillare intensamente. Dopo un po', tutto
		torna quietom, ma adesso tutti i poteri ", (artdi) self, " sono ripristinati.";
	],

	exausting[;
		"^L'energia emanata ", (artda) self, " @`e molto debole. Comprendi che
		non potrai usar", (itorthem) self," pi@`u.";
	];


!-------------------------------------------------------------------------------
! A magical battery is a magic device that provides an additional
! spellcasting ability to the player. Wands, magical rings, and the
! like will provide an additional spell casting source for mages: they
! are like scrolls, but they can be recharged by certain spells.
!
Class MagicBattery class MagicDevice
	with
	! sublcasses must provide a 'content' property
	charges 3,
	max_charges 3,
	prepared 0,		! readied?

	! subclases might overload this to print out a beautiful message
	! when a spell is cast from this object

	start_casting [;
		print (The) self, " irradia magia pura, mentre l'incantesimo viene rilasciato.^^";
		!two rows
	],

	recharge [;
	],


	has_spell[ spell i;
		for (i=0: i < self.#content/2: i++) {
			if (self.&content-->i == spell) rtrue;
		}
		rfalse;
	],

	! provides this function to signal to the MEMORY object that
	! this object can cast a spell
	can_cast[ sp;
		! this can cast if this object has been prepared, or
		! worn (if it's something you can put on), and only if
		! you know how to use that object

		if (self has known_about && self.has_spell(sp) && self.charges > 0 &&
			((self has clothing && self has worn) || self.prepared) )
			return MFS_BOOK_PRIORITY;
		return -1;
	],

	affect[flag;
		if (flag ~= SPELL_REVEAL) rfalse;
		if (self.describe_contents())
			give self known_about;
		rtrue;
	],

	describe_contents [ sp num memory i words;
		memory = GetMemory(actor);
		if (memory == nothing) {
			print "Solo un mago pu@`o usare ", (the) self, ".^";
			rfalse;
		}
		num = 0;

		objectloop(sp in memory)
			if (self.has_spell(sp))
				num ++;

		if (num == 0) {
			print "I poteri ", (artdi) self, " sono oscuri e arcani e, nonostante
			la tua conoscenza della magia, non sei in grado di penetrarli.^";
			rfalse;
		}
		words = self.#content/2;
		num = 0;

		for (i = 0: i < words: i++) {
			sp = self.&content-->i;
			if ( sp has known_about) {
				num++;
				if (i == 0)
					print (The) self, " pu essere usato per lanciare ";
				print (the) sp;
				if (i == words-1) print ".^";
				else if ( i == words-2) print " e ";
				else print ", ";
			}
		}

		if (num < words) {
			print "Oltre a questo, sei in grado di capire che ", (the) self, 
			" ha qualche potere nascosto che, nonostate i tuoi sforzi, non riesci a scoprire.^";
		}
		rtrue;
	],

	prepare [;
		if (self hasnt known_about)
			"Nonostante tutta la tua conoscenza, non riesci a capire come usare ", (the) self, ".";

		if (self.prepared == 1)
			print "Impugni ", (the) self, " di fronte a te.";
		if (self has clothing)
			"Non hai bisogno di preparare ", (the) self, "
				; basta indossarlo.";

		self.prepared = 1;
		"Adesso ", (the) self, " ", (isorare) self, " pront", (gan) self,
         " per scaricare il proprio potere!";
	],

	repose [;
		if (self.prepared == 1) {
			self.prepared = 0;
			"Riponi ", (the) self, ".";
		}
		"Ma ", (the) self, " non @`e mai stato preparato per lanciare gli incatesimi.";
	],

	invent[;
		if (inventory_stage==2 && self.prepared == 1)
			print " (pronto a lanciare gli incatesimi)";
		rfalse;
    ],

	cast[ sp tr rm ct;
		if (rm == 0) rm = location;
		if (ct == 0) ct = player;
		if (self in player) self.start_casting();
		if (tr == nothing) {
			sp.room = rm;
			sp.caster = ct;
			<FireSpell sp>;
		}
		else {
			<FireSpell sp tr>;
		}
		self.charges --;
		if (self.charges <= 0)
			if (self in player) self.exausting();
	],
	after [;
		Take:
			if (self.prepared == 1)
				self.prepared = 0;
			rfalse;
	];

!--------------------------------------------------------------------
! dummy spell, to illustrate the workout of the system
! No need for name, and never give it known_about
Spell object_spell
	with
	fire_to_all [;
		"^Questo oggetto @`e stato caricato con un incantesimo, che
		non far@`a nulla a parte stampare questa linea di testo.^";
	];

!---------------------------------------------------------------------
! A magic generator is a magic device with self mantained
! spellcasting abilities. It's like a "little mage" on its own.
! A magical generator has one or more spells, which can be actived
! saying one or more "power words", or doing some other key actions
! on the object ("> RUB THE MAGIC LAMP" i. e.).
! You'll have to create spells with fire_to_all property, or use
! existing spells, and list them in the "content" propriety
! of the generator. Then, for every entry in the "content" you'll add a
! entry (dictionary word described using ' instead of ")
! in the "activation" property; the word in the activation entry
! will launch the fire_to_all property of the corresponding spell.
! you may also provide a "charge" property if you want to limitate the
! use of the object.
! The activation words will be known only if the object will be given
! the known_about attribute; else they will appare as unknown verbs to
! the parser.
! If you want an action to trigger the spells in the object, you'll have
! to do the work by yourself...
!
class MagicGenerator class MagicDevice
	with
	content object_spell,		! the spell content array
	activation 'shazam',
	charges -1,					! infinite charges
	verbosity MFS_SPELLVMED,		! Medium verbosity in description

	! routine to be overloaded by subclasses when the player
	! discovers the magical powers of the object (called also
	! by magic revealing spells).

	describe_contents[i words;
		print "Puoi usare ", (the) self, " pronunciando ";
		words = self.#activation/2;
		for (i = 0: i < words: i++) {
			print "~", (address) self.&activation-->i,"~";
			if (i == words-1) print ".^";
			else if ( i == words-2) print " and ";
			else print ", ";
		}
	],

	affect[flag;
		if (flag ~= SPELL_REVEAL) rfalse;
		self.describe_contents();
		give self known_about;
		rtrue;
	],

	! indicates if all conditions to use this object are met;
	! by default, it's the player has it, it can use it, but if
	! this object has "clothing", it must also have "worn".
	usable[;
		if (self in player) {
			if (self hasnt clothing)
				rtrue;
			else if (self has worn)	rtrue;
		}
		rfalse;
	],

	! string to be printed if some conditions of using this
	! object are not met
	unable_to_cast [;
		if (self notin player)
			"@`E un peccatto che tu non abbia ", (the) self, " con te, al momento.";
		"Prima di poter usare i poteri ", (artdi) self, " devi indossar", (itorthem) self,".";
	],

	! if the object has no more powers...
	exausted[;
		print (The) self, " ha gi@`a liberato tutto il proprio potere, e le parole
		che hai pronunciato non sortiscono alcun effetto.^";
	],

	! tells if a given parameter is a power word for the object
	is_a_pword[pword;
		if (self.pword_number(pword)> -1) rtrue;
		rfalse;
	],

	! find the index of a power word
	pword_number[pword i;
		for(i = 0: i < self.#activation/2: i++) {
			if (self.&activation-->i == pword) {
				return i;
			}
		}
		return -1;
	],

	! cast a given spell
	cast[pword num spell;
		if (~~self.usable()) {
			self.unable_to_cast();
			rfalse;
		}
		if (self.charges == 0) {
			self.exausted();
			rfalse;
		}
		num = self.pword_number(pword);
		if (num == -1) rfalse;
		spell = self.&content-->num;
		spell.room = parent(parent(self));
		spell.caster = parent(self);
		<FireSpell spell>;
		if (self.charges == 1) self.exausting();
		if (self.charges > 0) self.charges --;
		rtrue;
	],

	! like is_a_pword but filtered by known_about
	usable_pword[ pword;
		if (self.is_a_pword(pword) && self has known_about) rtrue;
		rfalse;
	];

ifdef Weapon;
class WeaponBattery class Weapon MagicBattery;
class WeaponGenerator class Weapon MagicGenerator;
Endif;

!-----------------------------------------------------
! Cast the verb used as a magic word in a generator
!
[ CastGeneratorSub;
  if (BeforeRoutines())rtrue;
  the_object.cast(the_spell);
  return AfterRoutines();
];


Endif; ! ifndef MagicDevice

! end of MFS_MagicBattery
